##########################################################################
# biMM (1.0.0): An R package for a bivariate linear Mixed Model 
# DATE: 11-Nov-2016; update 18-Jan-2017 (See end of file for update log.)
# Copyright (C) 2016,2017 University of Helsinki
# Software written by Matti Pirinen with a contribution from Christian Benner.
# Contact: matti.pirinen@iki.fi
# LICENSE:
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <http://www.gnu.org/licenses/>.
#
##########################################################################
#' biMM: A package for a bivariate linear Mixed Model
#'
#' The biMM package provides a reference class \code{\link{biMM.data}}
#' for holding data and implementing methods for univariate
#' and bivariate variance component estimation of a
#' linear mixed model with one random effect.
#' See manual for \code{\link{biMM.data}} for details.
#' @author Matti Pirinen <matti.pirinen@iki.fi>
#' @docType package
#' @name biMM
NULL
#' @useDynLib biMM
#' @importFrom Rcpp evalCpp
#' @importFrom methods new

biMM.univariate.loglik <- function(x, Uty, d)
{
    #elements of x are heritability and total trait variance
    A <- x[2]*(1 + x[1]*(d-1)) 
    return(-0.5 * sum(log(A) + Uty^2 / A))
}

biMM.univariate.score <- function(x, Uty, d)
{
    #elements of x are heritability and total trait variance
    A = (1 + x[1]*(d-1))
    dx1 = 0.5 * sum( (d-1) / A * (Uty^2 / x[2] / A - 1) )
    dx2 = 0.5/x[2] * sum( Uty^2/A/x[2] - 1)
    return(c(dx1,dx2))
}

biMM.univariate.hessian <- function(x, Uty, d)
{
    #elements of x are heritability and total trait variance
    A = (1 + x[1]*(d-1))
    cross = -0.5/x[2]^2* sum( (d-1)* Uty^2/A^2)
    return(matrix(c(0.5 * sum( ((d-1) / A)^2 * (1- (2*Uty^2) / x[2] / A) ),
                 cross,cross,
                 0.5/x[2]^2*(sum(1-2*Uty^2/x[2]/A))),2,2))
}

biMM.bivariate.loglik.cov<-function(x,y1,y2,d)
{
  #6 elem of x are: vg1,ve1,vg2,ve2,vg12,ve12
  a1=x[1]*d+x[2]
  a2=x[3]*d+x[4]
  a12=x[5]*d+x[6]
  
  -0.5*sum(log(a1*a2-a12^2)+(a1*y2^2+a2*y1^2-2*a12*y1*y2)/(a1*a2-a12^2))
}

biMM.bivariate.loglik.cor<-function(x,y1,y2,d){
  #6 elem of x are: vg1,ve1,vg2,ve2,rg,re
  a1=x[1]*d+x[2]
  a2=x[3]*d+x[4]
  a12=x[5]*d*sqrt(x[1]*x[3])+x[6]*sqrt(x[2]*x[4])
  
  -0.5*sum(log(a1*a2-a12^2)+(a1*y2^2+a2*y1^2-2*a12*y1*y2)/(a1*a2-a12^2))
}

biMM.bivariate.score.cov<-function(x,y1,y2,d){
  #6 elem of x are: vg1,ve1,vg2,ve2,vg12,ve12
  a1=x[1]*d+x[2]
  a2=x[3]*d+x[4]
  a12=x[5]*d+x[6]
  C=a1*a2-a12^2
  B=a1*y2^2+a2*y1^2-2*a12*y1*y2

  dvg1=-0.5*sum(d*(C*(a2+y2^2)-a2*B)/C^2)
  dvg2=-0.5*sum(d*(C*(a1+y1^2)-a1*B)/C^2)
  dvg12=-1.0*sum(d*(-C*a12-C*y1*y2+B*a12)/C^2)
  dve1=-0.5*sum((C*(a2+y2^2)-a2*B)/C^2)
  dve2=-0.5*sum((C*(a1+y1^2)-a1*B)/C^2)
  dve12=-1.0*sum((-C*(a12+y1*y2)+a12*B)/C^2)

  c(dvg1,dve1,dvg2,dve2,dvg12,dve12)
}

biMM.bivariate.score.cor<-function(x,y1,y2,d){
  #6 elem of x are: vg1,ve1,vg2,ve2,rg,re
    sqrt.s = sqrt(x[1]*x[3])
    sqrt.t = sqrt(x[2]*x[4])

    a1 = x[1]*d+x[2]
    a2 = x[3]*d+x[4]
    a12 = x[5]*d*sqrt.s+x[6]*sqrt.t
    C = a1*a2-a12^2
    B = a1*y2^2+a2*y1^2-2*a12*y1*y2
    
    dvg1 = -0.5*sum(d/C^2*(C*(a2+y2^2)-a2*B+x[5]*sqrt(x[3]/x[1])*(B*a12-C*(y1*y2+a12))))
    dvg2 = -0.5*sum(d/C^2*(C*(a1+y1^2)-a1*B+x[5]*sqrt(x[1]/x[3])*(B*a12-C*(y1*y2+a12))))
    dve1 = -0.5*sum(1/C^2*(C*(a2+y2^2)-a2*B+x[6]*sqrt(x[4]/x[2])*(B*a12-C*(y1*y2+a12))))
    dve2 = -0.5*sum(1/C^2*(C*(a1+y1^2)-a1*B+x[6]*sqrt(x[2]/x[4])*(B*a12-C*(y1*y2+a12))))
    drg = -0.5*sum(2*d*sqrt.s/C^2*(a12*(B-C)-C*y1*y2))
    dre = -0.5*sum(2*sqrt.t/C^2*(a12*(B-C)-C*y1*y2))

    c(dvg1,dve1,dvg2,dve2,drg,dre)
}

biMM.bivariate.score.cor.2<-function(x,y1,y2,d){
  #This is old version that gives the same results as new above.  
  #The new has just been simplified a bit and made match the formulas in the write-up.
  #6 elem of x are: vg1,ve1,vg2,ve2,rg,re
  a1=x[1]*d+x[2]
  a2=x[3]*d+x[4]
  a12=x[5]*d*sqrt(x[1]*x[3])+x[6]*sqrt(x[2]*x[4])
  C=a1*a2-a12^2
  B=a1*y2^2+a2*y1^2-2*a12*y1*y2
  sdr=sqrt(x[3]/x[1])
  
  dvg1=-0.5*sum(d*(a2-a12*x[5]*sdr)/C)
  dvg1=dvg1-0.5*sum((d*(y2^2-y1*y2*x[5]*sdr)*C-d*(a2-a12*x[5]*sdr)*B)/C^2)

  dvg2=-0.5*sum(d*(a1-a12*x[5]/sdr)/C)
  dvg2=dvg2-0.5*sum((d*(y1^2-y1*y2*x[5]/sdr)*C-d*(a1-a12*x[5]/sdr)*B)/C^2)

  sdr=sqrt(x[4]/x[2])
  dve1=-0.5*sum((a2-a12*x[6]*sdr)/C)
  dve1=dve1-0.5*sum(((y2^2-y1*y2*x[6]*sdr)*C-(a2-a12*x[6]*sdr)*B)/C^2)

  dve2=-0.5*sum((a1-a12*x[6]/sdr)/C)
  dve2=dve2-0.5*sum(((y1^2-y1*y2*x[6]/sdr)*C-(a1-a12*x[6]/sdr)*B)/C^2)

  sdr=sqrt(x[1]*x[3])
  drg=sum((a12*d*sdr)/C)-0.5*sum(((-2*y1*y2*sdr*d)*C+2*a12*sdr*d*B)/C^2)
 
  sdr=sqrt(x[2]*x[4])
  dre=sum((a12*sdr)/C)-0.5*sum(((-2*y1*y2*sdr)*C+2*a12*sdr*B)/C^2)
 
  c(dvg1,dve1,dvg2,dve2,drg,dre)
}

#' Reference class for bivariate variance component analysis.
#'
#' @field d numeric vector, eigenvalues of relatedness (of Included).
#' @field Inculded logical vector, who are included in U and d.
#' @field IndID character vector, inds' IDs.
#' @field nInd integer, number of inds.
#' @field nTraits integer, number of traits.
#' @field relatedness numeric matrix, nInd x nInd GRM.
#' @field traits numeric matrix, nInd x nTraits.
#' @field trait.cor numeric matrix, correlation of traits.
#' @field U numeric matrix, eigenvectors of relatedness (of Included).
#' @section bivariate.heritability(--args above--):
#' 'trait.pairs', 2-column array naming the trait pairs in the order
#' they are to be analyzed.
#' 'allow.impute', for at most how many individuals allows at least one
#' trait value to be imputed;
#' 'allow.set.missing', for at most how many individuals allows the
#' trait values to be set missing;
#' 'traits.for.imputation', at most how many traits are used to impute missing
#' values. For imputation, chooses traits by their absolute correlation from 'trait.cor'
#' to the trait to be imputed. Value of 0 (DEFAULT) corresponds to the mean value imputation;
#' 'computation', if FALSE does not do the parameter estimation;
#' 'normalize' if FALSE does not scale the traits. Always mean centers the traits.
#' OUTPUT: data.frame with variance components for the grm component
#' ('Vg1', 'Vg2' and 'cor_g') and the environmental component ('Ve1', 'Ve2' and 'cor_e').
#' 'cor' is the total correlation.
#' 'new_eigen' is 1 if a eigen decomposition was done anew for the pair, otherwise 0.
#' @section excludeInd(--args above--):
#' Excludes individuals from biMM.data object based on IDs in 'exclIDs'.
#' If 'redo.cor' is FALSE, then does not recompute the trait.cor matrix after the exclusions.
#' @section excludeTraits(--args above--):
#' Excludes traits based on trait names given in 'exclTraits'.
#' @section heritability(--args above--):
#' 'trait.names', a vector naming the traits in the order
#' they are to be analyzed.
#' 'allow.impute', for at most how many individuals allows the
#' trait value to be imputed;
#' 'allow.set.missing', for at most how many individuals allows the
#' trait value to be set missing;
#' 'traits.for.imputation', at most how many traits are used to impute missing
#' values. For imputation, chooses traits by their absolute correlation from 'trait.cor'
#' to the trait to be imputed. Value of 0 (DEFAULT) corresponds to the mean value imputation;
#' 'computation', if FALSE does not do the parameter estimation;
#' 'normalize' if FALSE does not scale the traits. Always mean centers the traits.
#' OUTPUT: data.frame with the variance for the grm component
#' ('Vg') and the environmental component ('Ve'').
#' 'new_eigen' is 1 if a eigen decomposition was done anew for the pair, otherwise 0.
#' @section minimum.set.of.relateds.to.exclude(--args above--):
#' Finds an approximately minimum set of individuals that need to be excluded
#' so that at least one individual from each row of 'related.pairs' (2-column array of IDs)
#' is removed.
#' @section order.for.bivariate.heritability(--args above--):
#' Finds an order of trait pairs for 'bivariate.heritability()' aiming to optimize the
#' number of eigendecompositions and trait value imputation/missingness.
#' If 'trait.pairs' is NULL then considers all pairs of 'traits', otherwise considers
#' the pairs given as rows of 'trait.pairs';
#' If 'start.pair' is NULL then goes through all possible values for the first pair, otherwise
#' starts from the pair given in 'start.pair' (a vector of 2 trait names);
#' 'allow.impute' and 'allow.set.missing' are as for 'bivarite.heritability()' above.
#' @section order.for.heritability(--args above--):
#' Finds an order of traits for 'heritability()' aiming to optimize the
#' number of eigendecompositions and trait value imputation/missingness.
#' If 'trait.names' is NULL then considers all 'traits', otherwise considers
#' the traits given in 'trait.names';
#' If 'start.trait' is NULL then goes through all possible values for the first trait, otherwise
#' starts from 'start.trait';
#' 'allow.impute' and 'allow.set.missing' are as for 'heritability()' above.
#' @section read.traits.grm(--args above--):
#' Reads traits and/or genetic relationship matrix (grm).
#' File format for trait file: First two columns must match .grm.id file's columns 1-2
#' and trait file must have a header. For example, PLINK's ped files are OK,
#' but GCTA's trait files are not since they do not have a header.
#' File formats for .grm and grm.id is like in GCTA software
#' (but not accepting GCTA binary format).
#' Accepts also gzipped grm and the '.gz' ending may or may not be included in
#' input 'grm.filename'
#' However, for the ID file, '.id' is always added to the input string 'grm.filename',
#' even if 'grm.filename' ends in '.gz'.
#' So typically 'grm.filename' should not have '.gz' ending in it. 
#' @section relatedPairs(--args above--):
#' Lists pairs of individuals whose 'relatedness' value is above the 'threshold'.
#'
#' @section Examples:
#' x = biMM.data() ##Generate a new biMM.data object
#'
#' Exclude a minimum set of individuals 
#' so that no pair with r>0.8 remains:
#'
#' x$excludeInd(x$minimum.set.of.relateds.to.exclude(x$relatedPairs(0.8)))
#'
#'
#' Order traits for univariate heritability analysis:
#'
#' tr.ord = x$order.for.heritability(allow.impute=20, allow.set.missing=30)
#'
#'
#' Compute univariate heritabilities:
#'
#' x$heritability(tr.ord , allow.impute=20, allow.set.missing=30)
#'
#'
#' Order traits for bivariate heritability analysis:
#'
#' tr.ord = x$order.for.bivariate.heritability(allow.impute=20, allow.set.missing=30)
#'
#'
#' Compute bivariate heritabilities:
#'
#' x$bivariate.heritability(tr.ord , allow.impute=20, allow.set.missing=30)
#'
#' @export biMM.data
#' @exportClass biMM.data

biMM.data <- methods::setRefClass("biMM.data", 
    fields = list(
        relatedness = 'matrix', #could also be just lower triangle if space was issue
        IndID = 'character',
        traits = 'matrix',
        trait.cor = 'matrix',
        nInd = 'integer',
        nTraits = 'integer',
        U = 'matrix',
        d = 'numeric',
        Included = 'logical'
    ),
    methods = list(
        
        printTraits = function(){
            " "#"Prints names of the traits."
            return(colnames(traits))
        },
        
        printInds = function(){
            " "#"Prints IDs of the individuals."
            return(IndID)
        },

        excludeInd = function(exclIDs, redo.cor = TRUE){
            " "#"Excludes data on the given IDs."
            ind = (IndID %in% exclIDs)
            cat(paste("Excludes",sum(ind),"individuals.\n"))
            if(sum(ind)>0){
                if(nrow(relatedness) > 0) relatedness <<- relatedness[!ind,!ind]
                if(nrow(traits) > 0) traits <<- traits[!ind,]
                if(redo.cor & (nrow(traits) > 0)) trait.cor <<- cor(traits, method = "pearson", use = "pairwise.complete.obs")
                IndID <<- IndID[!ind]
                nInd <<- length(IndID)
                U <<- matrix(NA,0,0)
                d <<- numeric(0)
                Included <<- rep(FALSE,nInd)
            }
            cat(paste("Remaining",nInd,"individuals.\n"))
        },

        excludeTraits = function(exclTraits){
            " "#"Excludes the given traits."
            ind = (colnames(traits) %in% exclTraits)
            cat(paste("Excludes",sum(ind),"traits.\n"))
            if(sum(ind)>0){
                traits <<- traits[,!ind]
                trait.cor <<- trait.cor[!ind,!ind]
                nTraits <<- ncol(traits)
            }
            cat(paste("Remaining",nTraits,"traits.\n"))
        },

        relatedPairs = function(threshold){
            " "#"Outputs pairs related above the threshold."
            cat(paste0("Finds related above threshold ",threshold,".\n"))
            rel = which(relatedness>threshold, arr.ind=TRUE)
            rel[rel[,1] != rel[,2],]
            if(nrow(rel)>0){
                rel = matrix(rel[rel[,1] > rel[,2],],ncol=2)
                rel = data.frame( ind1 = IndID[as.numeric(rel[,1])] , ind2 = IndID[as.numeric(rel[,2])],
                                 r = relatedness[rel[,1]+(rel[,2]-1)*nInd],stringsAsFactors = FALSE)
                cat(paste("Found",nrow(rel),"pairs of related individuals.\n"))
            }else{
                cat(paste("Found no pairs of related individuals.\n"))}
            return (rel)
        },

        minimum.set.of.relateds.to.exclude = function(related.pairs){
            " "#"Outputs a set of IDs to exclude in order to remove at least one of each pair given."
            cat(paste("Finds an approximate minimum vertex cover of relatedness graph.\n"))
            ind = unique(c(related.pairs[,1],related.pairs[,2]))
            nind = length(ind)
            cat(paste("Starts with",nrow(related.pairs),"pairs involving",nind,"unique individuals.\n"))
            deg = rep(0,nind)
            rel = cbind(match(related.pairs[,1],ind),match(related.pairs[,2],ind))
            neigh = list()
            #Make list of neighbours and count degrees
            for(ii in 1:nrow(rel)){
                k = c(rel[ii,1],rel[ii,2])
                for(jj in 1:2){
                    deg[k[jj]] = deg[k[jj]]+1
                    if(deg[k[jj]]==1){neigh[[k[jj]]] = c(k[3-jj])} #initialize
                    else{ neigh[[ k[jj] ]] = c(neigh[[ k[jj] ]], k[3-jj]) }
                }
            }
            excl=rep(NA,nind)
            i.excl = 1
            while(max(deg)>0){
                ii = which.max(deg)
                for(jj in neigh[[ii]]){
                    deg[jj] = deg[jj]-1
                    neigh[[jj]] = neigh[[jj]][neigh[[jj]] != ii]
                }
                deg[ii] = 0
                excl[i.excl]=ii
                i.excl = i.excl + 1
            }
            cat(paste("Suggests excluding",i.excl-1,"individuals.\n"))
            return (ind[excl[1:(i.excl-1)]])
        },

        order.for.heritability = function(trait.names=NULL, start.trait=NULL, allow.impute=0, allow.set.missing=0){
            " "#"Outputs an order of traits to be used when calling heritability()."
        #If No trait.names specified, then all traits are considered
        #If no start.trait specified, then looping over all traits as start trait and choosing the one
        # that gives the smallest number of eigen decompositions given the allow.impute and allow.set.missing parameters
        
            if(is.null(trait.names)) trait.names=colnames(traits)
            trait.names = unique(trait.names)
            ind = trait.names %in% colnames(traits)
            cat(paste("Found",sum(ind),"traits from trait data.\n"))
            if(!all(ind)) {
                cat(paste("Did not find these traits:\n"))
                print(trait.name[!ind])
            }
            avail = !is.na(traits[,trait.names])
            
            if(!is.null(start.trait)){
                start.i = which(trait.names == start.trait)[1]
                if( length(start.i)==0 ) stop("Did not find start.pair from trait.pairs")
            }else{
                sum.hh = matrix(NA,ncol=3,nrow=length(trait.names),
                                dimnames=list(NULL,c("new_eigen","median imputed","median ignored")))
                cat("Searching for the best starting point:\n")
                for(ii in 1:length(trait.names)){
                    ord = .Call('HamiltonianPathForColumns', avail_ = avail, start_i_ = ii, PACKAGE = 'biMM')
                    hh = .self$heritability(trait.names[ord] , allow.impute=allow.impute,
                                            allow.set.missing=allow.set.missing, computation=FALSE)
                    sum.hh[ii,]=c(sum(hh[,"new_eigen"]),median(hh[,"N_imputed"]),median(hh[,"N_ignored"]))
                    cat("\r", round(ii/length(trait.names)*100,1),"% done.") 
                    flush.console()
                }
                cat("\n")
                print(data.frame(first_trait=trait.names,sum.hh))
                ii = which(sum.hh[,1]==min(sum.hh[,1])) #First, order by number of eigen decompositions
                start.i = ii[which(sum.hh[ii,2] == min(sum.hh[ii,2]))[1]] #Second, order by number of imputed values
            }
            cat(paste("Starts from",trait.names[start.i],"\n"))
            ord = .Call('HamiltonianPathForColumns', avail_ = avail, start_i_ = start.i, PACKAGE = 'biMM')
            return(trait.names[ord])
        },

        order.for.bivariate.heritability = function(trait.pairs=NULL, start.pair=NULL, allow.impute=0, allow.set.missing=0){
             " "#"Outputs an order of pairs of traits to be used when calling bivariate.heritability()."
             if(is.null(trait.pairs)) {
                 ii1 = 1 + floor( (0:(nTraits^2-1)) / nTraits )
                 ii2 = 1 + ( (0:(nTraits^2-1)) %% nTraits)
                 ii.inc = (ii2<ii1) #include only those where col index is < row index
                 trait.pairs=cbind(colnames(traits)[ii1[ii.inc]], colnames(traits)[ii2[ii.inc]])
             }else{
                 if(is.vector(trait.pairs)) trait.pairs = matrix(trait.pairs,byrow=TRUE, ncol=2)
                 if(ncol(trait.pairs)!=2) stop("trait.pairs should have exactly two columns")
                 ind = apply(matrix(trait.pairs %in% colnames(traits),byrow=FALSE,ncol=2),1,all)
                 trait.pairs = matrix(trait.pairs[ind,],ncol=2)
                 if(length(trait.pairs) < 2) stop("None of the given trait pairs found.")
                 trait.pairs = matrix(trait.pairs[!duplicated(paste(trait.pairs[,1],trait.pairs[,2],sep="__")),],ncol=2)
             }
             cat(paste("Found",nrow(trait.pairs),"trait pairs from trait data.\n"))
             if(nrow(trait.pairs)==1) return(trait.pairs) #The only possible ordering
             
             avail = !is.na(traits[,trait.pairs[,1]] + traits[,trait.pairs[,2]])
             
             if(!is.null(start.pair)){
                 start.i = which((trait.pairs[,1] == start.pair[1] & trait.pairs[,2] == start.pair[2]) |
                                 (trait.pairs[,2] == start.pair[1] & trait.pairs[,1] == start.pair[2]))[1]
                 if( length(start.i)==0 ) stop("Did not find start.pair from trait.pairs")
             }else{
                 sum.hh = matrix(NA,ncol=3,nrow=nrow(trait.pairs),
                                 dimnames=list(NULL,c("new_eigen","median imputed","median ignored")))
                 cat("Searching for the best starting point:\n")
                 for(ii in 1:nrow(trait.pairs)){
                     ord = .Call('HamiltonianPathForColumns', avail_ = avail, start_i_ = ii, PACKAGE = 'biMM')
                     hh = .self$bivariate.heritability(trait.pairs[ord,] , allow.impute=allow.impute,
                                                       allow.set.missing=allow.set.missing, computation=FALSE)
                     sum.hh[ii,]=c(sum(hh[,"new_eigen"]),median(hh[,"N_imputed"]),median(hh[,"N_ignored"]))
                     cat("\r", round(ii/nrow(trait.pairs)*100,1),"% done.") 
                     flush.console()
                 }
                 cat("\n")
                 print(data.frame(first_pair_1 = trait.pairs[,1],first_pair_2 = trait.pairs[,2],sum.hh))
                 ii = which(sum.hh[,1]==min(sum.hh[,1])) #First, order by number of eigen decompositions
                 start.i = ii[which(sum.hh[ii,2] == min(sum.hh[ii,2]))[1]] #Second, order by number of imputed values
             }
             cat(paste("Starts from pair",paste(trait.pairs[start.i,],collapse=" -- "),"\n"))
             ord = .Call('HamiltonianPathForColumns', avail_ = avail, start_i_ = start.i, PACKAGE = 'biMM')
             return(trait.pairs[ord,])
        },

        read.traits.grm = function( trait.filename=NULL, grm.filename=NULL, id.col=2, merge.ids=FALSE){
            " "#"Reads traits and/or genetic relationship matrix (grm)."

            #File format for trait file:
            #First two columns must match .grm.id file's columns 1-2
            # and trait file must have header. 
            #For example, PLINK's ped files are OK, but this is different from
            #GCTA's pheno file that does not have a header.
            #If merge.ids == TRUE, then combines ids in columns 1 and 2 by an underscore
            #to make IndID for downstream analyses.
            #If merge.ids == FALSE then uses as IndID column 'id.col' from the trait file (either 1 or 2).
            
            #File formats for .grm and grm.id is like GCTA (but not accepting GCTA binary format)
            #Accepts also gzipped grm and the '.gz' ending may or may not be included in input 'grm.filename'
            #However, for the ID file, '.id' is always added to the input 'grm.filename',
            # even if 'grm.filename' ends in '.gz'
            #So usually 'grm.filename' should not have '.gz' ending in it. 

            if( !merge.ids & !(id.col %in% c(1,2)) ) stop("ERROR id.col must be either 1 or 2.")
            
            #READ TRAITS
            if(!is.null(trait.filename)){
                
                if(file.access(trait.filename, mode = 4)==(-1))
                    stop(paste("ERROR: Can't access trait file:",trait.filename));
                cat(paste("Reading traits from",trait.filename,"\n"))
                if(merge.ids) cat(paste("Merging ids from columns 1 and 2 by separator \"_\".\n"))
                else cat(paste0("Using ID column ",id.col,".\n"))
                pheno = data.table::fread(trait.filename,header=TRUE,data.table=FALSE)
                if(merge.ids) {tr.ids = paste(pheno[,1],pheno[,2],sep="_")}
                else{ tr.ids = as.character(pheno[,id.col]) }
                if(any(duplicated(tr.ids))) {
                    k = which(duplicated(tr.ids))[1]
                    k = which(tr.ids == tr.ids[k])
                    stop(paste0("ERROR: IDs not unique; Individuals ",paste(k,collapse=", ")," have the same ID: ",tr.ids[k[1]],"\n",
                                "   You can either choose column 1 or 2 by 'id.col' or\n",
                                "   merge the two columns to one ID by setting merge.ids=TRUE."))
                }
                #ids = pheno[,1:2]
                if( (length(IndID)>0) & !all(tr.ids == IndID) )
                    stop("ERROR: The IDs read from ",trait.filename,
                         " and existing IndIDs of the biMM.data object do not match.\n",
                         "  Maybe you want to crete a new biMM.data object?") 
                #IndID <<- tr.ids #Needed in case IndID was empty
                
                pheno = pheno[,3:ncol(pheno)]
                remove.ind = !(sapply(pheno,typeof) %in% c("double","integer"))
                if(any(remove.ind)){
                    cat(paste("Removing following",sum(remove.ind),"columns since they are not numeric:\n"))
                    print(paste(names(pheno)[remove.ind]))
                }
                pheno = as.matrix(pheno[,!remove.ind])
                cat(paste("Read data for",length(tr.ids),"individuals on",ncol(pheno),"traits.\n"))
            }
            
            #READ RELATEDNESS MATRIX
            if(!is.null(grm.filename)){
                
                id.file = paste0(grm.filename,".id")
                if(file.access(id.file, mode = 4)==(-1))
                    stop(paste("ERROR: Can't access ID file:",id.file));
                cat(paste("Reading IDs from",id.file,"\n"))
                grm.ids = data.table::fread(id.file,header=FALSE,data.table=FALSE)
                if(merge.ids) {grm.ids = paste(grm.ids[,1],grm.ids[,2],sep="_")}
                else{ grm.ids = as.character(grm.ids[,id.col]) }
                    
                if(!is.null(trait.filename)){
                    if(!all(tr.ids == grm.ids)){
                        k = which(tr.ids != grm.ids)[1]
                        if(merge.ids){
                            stop("ERROR: IDs from ",id.file," and\n",
                                 trait.filename," do not match when merge.ids=TRUE.\n",
                                 "Ind: ",k," Ids:",grm.ids[k]," ",tr.ids[k]) }
                        else{
                            stop("ERROR: IDs from ",id.file," and\n",
                                 trait.filename," do not match at column ",id.col,".\n",
                                 "Ind: ",k," Ids:",grm.ids[k]," ",tr.ids[k]) }
                    }
                }
                if( (length(IndID)>0) & !all(grm.ids == IndID) ){
                    stop("ERROR: Ids from ",id.file,
                         " and existing IDs of the biMM.data object do not match.\n",
                         "  Maybe you want to crete a new biMM.data object?") }
                
                if(file.access(grm.filename, mode = 4)==(-1)) {
                    cat(paste0("Can't find grm file: ",grm.filename,".\n  ---> Trying .gz instead.\n")); 
                    grm.filename = paste0(grm.filename,".gz")
                    if(file.access(grm.filename, mode = 4)==(-1))
                        stop(paste("Can't find gzipped grm file:",grm.filename,
                                   "\n Either .grm file or .grm.gz file must be accessible."));
                }
                
                cat(paste("Reading grm from",grm.filename,"\n"))
                if(regexpr("(.gz)$",grm.filename) >= 0)
                    Rlist <- data.table::fread(paste("gunzip -c",grm.filename), header=FALSE,data.table=FALSE)
                else
                    Rlist <- data.table::fread(grm.filename, header=FALSE,data.table=FALSE)

                nInd <<- length(grm.ids) # Won't abort anymore after this point: can assign to object with <<- from now on
                relatedness <<- matrix(NA,nInd,nInd)
                relatedness[Rlist[,1] + nInd*(Rlist[,2]-1)] <<- Rlist[,4]
                relatedness[Rlist[,2] + nInd*(Rlist[,1]-1)] <<- Rlist[,4]
                cat(paste("Read relatedness matrix for",nInd,"individuals.\n"))
                if(any(is.na(relatedness))){
                    k = as.numeric(which(is.na(relatedness),arr.ind=T)[1,])
                    cat(paste0("\nLeft empty ",sum(is.na(relatedness))," elements from the matrix, e.g., element (",
                               k[1],",",k[2],").\n"))
                    cat(paste("Replacing them with 0.\nMAKE SURE you do this ON PURPOSE.\n\n"))
                    relatedness[is.na(relatedness)] <<- 0
                }
                
                #OLD version using an Rcpp to read in relatedness.
                #    data.table::fread above is even faster  
                #relatedness <<- .Call('list2matrix', vec_ = Rlist[,4], PACKAGE = 'biMM')
       
            }
            if(!is.null(trait.filename)){ #assign these only here at the end so that if aborts earlier, no assignements made
                traits <<- pheno
                nTraits <<- ncol(pheno)
                trait.cor <<- cor(traits, method = "pearson", use = "pairwise.complete.obs")
                IndID <<- tr.ids
            }else{ IndID <<- grm.ids }
            nInd <<- length(IndID)
            U <<- matrix(NA,0,0)
            d <<- numeric(0)
            Included <<- rep(FALSE,nInd)
        },

        heritability = function(trait.names , allow.impute=0, allow.set.missing=0,
                                traits.for.imputation=0,computation=TRUE, normalize=TRUE){
            " "#"Calculates heritabilities of the traits in the order given."
            pr = computation
            load.U.d=save.U.d=NULL #Could save and load data for debugging
            if(!is.vector(trait.names)) stop("'trait.names' should be a vector of trait names")
            traits.for.imputation = round(traits.for.imputation)
            if(traits.for.imputation < 0) stop("'traits.for.imputation' should be a non-negative integer")
            trait.names = trait.names[trait.names %in% colnames(traits)]
            trait.ind = match(trait.names,colnames(traits))
            n.traits = length(trait.ind)
            if( n.traits == 0 ) stop("Did not find any of the trait.names from trait data columns.")
            if(pr) cat(paste("Estimates heritability  of",n.traits,"traits.\n"))
            res = data.frame(matrix(NA,ncol=7,nrow=n.traits,
                                    dimnames=list(NULL,c("Trait","N","N_imputed","N_ignored","Herit","SE","new_eigen"))))
            
            if(!computation) Included.orig = Included
            nBFGS=0;nNM=0;
            for(ii in 1:n.traits){
                ii.t = trait.ind[ii]
                if(pr) cat(paste("Trait:",colnames(traits)[ii.t],"\n"))
                
                imp.ind = sum(is.na(traits[,ii.t]) & Included)
                mis.ind = sum(!is.na(traits[,ii.t]) & !Included)
                if(pr) cat(paste0("Would need to impute ",imp.ind," and set missing ",mis.ind," individuals...\n"))
                if( imp.ind <= allow.impute & mis.ind <= allow.set.missing & sum(Included)>0){
                    if(pr) cat("...OK. --- No eigen decomposition.\n") 
                    y = as.numeric(scale(traits[Included,ii.t],scale=normalize)) #remove values not included in existing U and d
                    y[is.na(y)] = 0 #mean impute missing values compared to existing U and d
                    if(computation & imp.ind > 0 & traits.for.imputation > 0){ #will try to use some other traits for the imputation
                        ord.cor = rev(order(abs(trait.cor[ii.t,])))[-1] #other traits ordered descending w.r.t abs cor
                        Y = scale(traits[Included,], scale=TRUE) #Y is always normalized -- y may be scaled back if normalize==FALSE
                        for( ii.imp in which(is.na(Y[,ii.t])) ){
                            use.traits = ord.cor[!is.na(Y[ii.imp, ord.cor])]
                            if(length(use.traits) > 0){ #otherwise nothing to do since y[ii.imp] is already 0 which is the correct imputed value
                                use.traits = use.traits[1:min(c(traits.for.imputation,length(use.traits)))]
                                y[ii.imp] = t(trait.cor[ii.t,use.traits]) %*% solve(trait.cor[use.traits,use.traits],Y[ii.imp,use.traits])
                            }
                        }
                        if(!normalize) y[is.na(Y[,ii.t])] = sd(traits[,ii.t],na.rm=TRUE)*y[is.na(Y[,ii.t])] #by default imputed values are normalized
                    }
                    res[ii,7]=0
                    eff.ind = sum(!is.na(traits[,ii.t])) - mis.ind
                }else{ #Needs to do eigen decomposition
                    ind = !is.na(traits[,ii.t])
                    if(pr) cat(paste("...too much. --- Does an eigen decomposition for",sum(ind),"individuals.\n"))
                    y = as.numeric(scale(traits[ind,ii.t],scale=normalize))
                    Included <<- ind
                    eff.ind = sum(ind)
                    res[ii,7]=1
                    
                    if(computation){
                        R = relatedness[ind,ind]
                        do.eigen = TRUE #will do eigen if cannot read it from load.U.d
                        if(!is.null(load.U.d) ){
                            U.d.filename = paste0(load.U.d,"_",colnames(traits)[ii.t],"_U_d.Rdata")
                            if(file.access(U.d.filename, mode = 4)==(-1)){
                                cat(paste0("Can't access pre-computed U.d file: ",U.d.filename,"\n"))
                            }
                            else{
                                cat(paste0("Reads in pre-computed U.d file: ",U.d.filename,"\n"))
                                load(U.d.filename)
                                if(length(d) != eff.ind) {
                                    cat(paste0("But pre-computed U.d file: ",U.d.filename," does not match nInd=",nInd,"\n"))
                                }else{ do.eigen = FALSE}
                            }
                        }
                        if(do.eigen){
                            eig = eigen(R,sym=TRUE)
                            U <<- eig$vectors
                            d <<- eig$values
                            
                            if( !is.null(save.U.d) ){
                                U.d.filename = paste0(save.U.d,"_",colnames(traits)[ii.t],"_U_d.Rdata")
                                cat(paste0("Saves U and d to file: ",U.d.filename,"\n"))
                                save(U,d,file = U.d.filename)
                            }
                        }
                        imp.ind = 0
                        mis.ind = 0
                    }
                }
                res[ii,1] = c(colnames(traits)[ii.t])
                res[ii,2:6] = as.numeric(c(eff.ind,imp.ind,mis.ind,NA,NA))
                if(computation){
                    Uty = crossprod(U,y)
                    op.NM = optim( c(0.5,var(y)), fn = biMM.univariate.loglik, Uty = Uty, d = d,
                                  method = 'Nelder-Mead',control=list(fnscale=-1))
                    op.BFGS =  tryCatch({ optim( op.NM$par, fn = biMM.univariate.loglik, gr = biMM.univariate.score,
                                                Uty = Uty, d = d, method = 'BFGS', control=list(fnscale=-1) )},
                                        error = function(err){cat("Not able to optimize by BFGS.\n");return(NA)})
                    if(!is.na(op.BFGS[1]) & op.BFGS$convergence == 0 & all(op.BFGS$par >= 0)){
                        op = op.BFGS; nBFGS=nBFGS+1}
                    else {op = op.NM; nNM=nNM+1}
                    hessian = biMM.univariate.hessian(op$par,Uty,d)
                    var.matrix = tryCatch({ solve(-hessian)}, error = function(err){cat("Not able to invert Hessian.\n");return(NA)})
                    #print(op)
                    #print(hessian)
                    SEs = rep(NA,length(op$par))
                    if(!is.na(var.matrix) & !is.null(var.matrix) & !any(is.nan(var.matrix))) SEs = sqrt(diag(var.matrix))*sqrt(length(y)/eff.ind)
                    res[ii,5:6] = c(op$par[1],SEs[1])
                    if(res[ii,5]<0) res[ii,5]=0
                    if(res[ii,5]>1) res[ii,5]=1
                }
                #if(pr) print(res[ii,])
            }
            
            if(!computation) Included <<- Included.orig
            if(computation) cat(paste("Optimized with BFGS:",nBFGS," NM:",nNM,"\n"))
            return(res)       
        },

        bivariate.heritability = function(trait.pairs , allow.impute=0, allow.set.missing=0,
                                          traits.for.imputation=0, computation=TRUE, normalize=TRUE){
            " "#"Calculates bivariate variance components of the trait pairs in the order given."
            pr = computation
            if(is.vector(trait.pairs)) trait.pairs = matrix(trait.pairs, byrow=TRUE, ncol=2)
            if(!ncol(trait.pairs)==2) stop("'trait.pairs' should have two columns of trait names")
            trait.pairs = matrix(trait.pairs[trait.pairs[,1] %in% colnames(traits) & trait.pairs[,2] %in% colnames(traits),],ncol=2)
            n.pairs = nrow(trait.pairs)
            traits.for.imputation = round(traits.for.imputation)
            if(traits.for.imputation < 0) stop("'traits.for.imputation' should be a non-negative integer")
            if( n.pairs == 0 ) stop("Did not find any of the trait.pairs from trait data columns.")
            if(pr) cat(paste("Estimates bivariate model of",n.pairs,"pairs.")) 
            res = data.frame(matrix(NA,ncol=20,nrow=n.pairs,
                                    dimnames=list(NULL,c("Trait1","Trait2","N","N_imputed","N_ignored",
                                                         "Vg1","Vg1_SE","Ve1","Ve1_SE","Vg2","Vg2_SE","Ve2","Ve2_SE",
                                                         "cor_g","cor_g_SE","cor_e","cor_e_SE","cor","cor_SE",
                                                         "new_eigen"))))
            
            if(!computation) Included.orig = Included 
            nBFGS=0;nNM=0;
            for(ii in 1:n.pairs){
                ts = as.vector(trait.pairs[ii,]) 
                if(pr) cat(paste("Traits:",ts[1],ts[2],"\n"))
                
                imp.ind = sum( (is.na(traits[,ts[1]]) | is.na(traits[,ts[2]])) & Included) #would impute at least one trait for these individuals 
                mis.ind = sum(!is.na(traits[,ts[1]]) & !is.na(traits[,ts[2]]) & !Included) #would drop this many inds with COMPLETE data 
                if(pr) cat(paste0("Would need to impute for ",imp.ind," individuals and set missing values for ",mis.ind," individuals ...\n"))
                do.eigen = TRUE #by default
                if( imp.ind <= allow.impute & mis.ind <= allow.set.missing & sum(Included)>0){
                    y = traits[Included,ts] #remove values not included in existing U and d 
                    y[apply(is.na(y[,ts]),1,any),] = c(NA,NA) #If one trait missing make the other trait missing as well
                    min.sd = min(apply(y,2,function(x){sd(x,na.rm=TRUE)}))
                    if(!is.na(min.sd) & min.sd > 0){ #otherwise one trait is constant and correlation doesn't make sense -- so try doing eigen
                        if(pr) cat("...OK. --- No eigen decomposition.\n")
                        do.eigen = FALSE 
                        y = scale(y,scale=normalize)
                        y[is.na(y)] = 0 #mean impute missing values compared to existing U and d
                        if(computation & imp.ind > 0 & traits.for.imputation > 0){ #will try to use some other traits for the imputation
                            Y = scale(traits[Included,], scale=TRUE) #Y is always normalized -- y may be scaled back if normalize==FALSE
                            for(imp.trait in 1:2){
                                ii.t = which(colnames(traits) == ts[imp.trait])
                                ord.cor = rev(order(abs(trait.cor[ii.t,]))) #traits ordered descending w.r.t abs cor
                                ord.cor = ord.cor[ord.cor != ii.t] #remove the trait to be imputed from possible traits to be used for imputation
                                #ord.cor = ord.cor[!(ord.cor %in% which(colnames(traits) %in% ts))] #remove BOTH traits from traits used for imputation
                                
                                for( ii.imp in which(apply(is.na(Y[,ts]),1,any))){#if either one is missing impute BOTH #which(is.na(Y[,ii.t])) ){
                                    use.traits = ord.cor[!is.na(Y[ii.imp, ord.cor])]
                                    if(length(use.traits) > 0){ #otherwise nothing to do since y[ii.imp] is already 0 which is the correct imputed value
                                        use.traits = use.traits[1:min(c(traits.for.imputation,length(use.traits)))]
                                        y[ii.imp,imp.trait] = t(trait.cor[ii.t,use.traits]) %*% solve(trait.cor[use.traits,use.traits],Y[ii.imp,use.traits])
                                    }
                                }
                                if(!normalize) y[is.na(Y[,ii.t]),imp.trait] = sd(traits[,ii.t],na.rm=TRUE)*y[is.na(Y[,ii.t]),imp.trait] #by default imputed values are normalized
                            }
                        }
                        res[ii,"new_eigen"]=0
                        eff.ind = sum(!is.na(traits[Included,ts[1]]) & !is.na(traits[Included,ts[2]]))
                    }else{
                        if(pr) cat("...would be OK, EXCEPT that one trait is constant. Tries eigen decomposition.\n")
                        eff.ind = 0
                        do.eigen = TRUE
                    }
                }
                if(do.eigen){ #Needs to do eigen decomposition
                    ind = !(apply(is.na(traits[,ts]),1,any))
                    y = scale(traits[ind,ts],scale=normalize)
                    min.sd = min(apply(y,2,function(x){sd(x,na.rm=TRUE)}))
                    if(!is.na(min.sd) & min.sd > 0){ #if this fails,
                                        #then one trait is constant and corr below is NA and there won't be computation
                        if(pr) cat(paste("...too much. --- Does an eigen decomposition for",sum(ind),"individuals.\n"))
                        Included <<- ind
                        eff.ind = sum(ind)
                        res[ii,"new_eigen"]=1
                        if(computation){
                            R = relatedness[ind,ind]
                            eig = eigen(R,sym=TRUE)
                            U <<- eig$vectors
                            d <<- eig$values
                            imp.ind = 0
                            mis.ind = 0
                        }
                    }else{ eff.ind=0 }
                }
                corr = as.numeric(cor(y[,1],y[,2]))
                res[ii,3:(ncol(res)-1)] = c(eff.ind,imp.ind,mis.ind,rep(NA,12),corr,
                                            sqrt(1-corr^2)/(sqrt(nrow(y)-2)))
                res[ii,1:2]=ts
                if(computation & is.finite(corr) & !is.nan(corr)){
                    start.val = rep(NA,6)
                    Uty = crossprod(U,y)
                    #save(Uty,d,file="Uty_d.Rdata") #for debugging the optimization
                    #initialize by univariate results:
                    for(jj in 1:2){
                        op = optim( c(0.5,1), fn = biMM.univariate.loglik, Uty = Uty[,jj], d = d,
                                   method = 'Nelder-Mead',control=list(fnscale=-1))
                        #print(op)
                        op.BFGS = tryCatch({optim( op$par, fn = biMM.univariate.loglik, gr = biMM.univariate.score,
                                                  Uty = Uty[,jj], d = d, method = 'BFGS',control=list(fnscale=-1))},
                                           error=function(err){return(NA)})
                        #print(op.BFGS)
                        if(!is.na(op.BFGS[1])) op = op.BFGS
                        start.val[(2*jj-1):(2*jj)] = c(op$par[1],1-op$par[1])*op$par[2]
                    }
                    for(i.st in 1:4) if(start.val[i.st]<0.01) start.val[i.st]=0.01	
                    start.val[5:6] = corr*c(sqrt(prod(start.val[c(1,3)])),sqrt(prod(start.val[c(2,4)]))) #initialize by overall correlation

                #Do the bivariate model:
		#First more stable parameterisation using covariances
		#Then Nelder-Mead using correlations
		#Then BFGS using correlations
		#If BFGS fails, report results from Nelder-Mead, otherwise report BFGS
                #print(start.val)
                    op = optim( start.val, fn = biMM.bivariate.loglik.cov, y1 = Uty[,1], y2 = Uty[,2], d = d,
                               method = 'Nelder-Mead',control=list(fnscale=-1,maxit=10000))
                    start.val = op$par;
                    for(i.st in 1:4) if(start.val[i.st]<0.01) start.val[i.st]=0.01		
                    start.val[5] = start.val[5]/sqrt(start.val[1]*start.val[3])
                    start.val[6] = start.val[6]/sqrt(start.val[2]*start.val[4])
                    
                    for(i.st in 5:6) {
                        if(start.val[i.st]<(-0.99)) start.val[i.st]=(-0.95)
                        if(start.val[i.st]>0.99) start.val[i.st]=0.95
                    }
                    #print(start.val)
                    op.NM = tryCatch({optim( start.val,  fn = biMM.bivariate.loglik.cor,
                                            y1 = Uty[,1], y2 = Uty[,2], d = d, method = 'Nelder-Mead',
                                            control=list(fnscale=-1,maxit=10000),hessian=TRUE)},
                                     error=function(err){cat("ERROR: NM not able to do Hessian.\n");
                                         return(optim( start.val,  fn = biMM.bivariate.loglik.cor,
                                                      y1 = Uty[,1], y2 = Uty[,2], d = d, method = 'Nelder-Mead',
                                                      control=list(fnscale=-1,maxit=10000),hessian=FALSE))})
                    #print(op.NM$par)
                    op.BFGS =  tryCatch({optim( op.NM$par,  fn = biMM.bivariate.loglik.cor, gr = biMM.bivariate.score.cor,
                                               y1 = Uty[,1], y2 = Uty[,2], d = d, method = 'BFGS',hessian=TRUE,
                                               control=list(fnscale=-1,maxit=10000))},
                                        error = function(err){cat("ERROR: Not able to optimize by BFGS.\n");return(NA)})
                    #print(op.BFGS)
                    if(!is.na(op.BFGS[1]) & op.BFGS$convergence == 0 & all(op.BFGS$par[1:4] >= 0) & 
                       all(op.BFGS$par[5:6] >= (-1)) &  all(op.BFGS$par[5:6] <= 1)){
                        op = op.BFGS; nBFGS=nBFGS+1}
                    else {op = op.NM; nNM=nNM+1}
                    SEs = rep(NA,length(op$par))
                    if(!is.null(op$hessian) & !any(is.nan(op$hessian))) SEs = sqrt(diag(-solve(op$hessian)))*sqrt(nrow(y)/eff.ind)
                    res[ii,6:17] = as.vector(matrix(c(op$par,SEs),byrow=TRUE,nrow=2))
                }
                #if(pr) print(res[ii,])
            }
            if(!computation) Included <<- Included.orig
            if(computation) cat(paste("Optimized with BFGS:",nBFGS," NM:",nNM,"\n"))
            return(res)       
        }
    )
 )

#UPDATE LOG
#18-Jan-2017: (Small fixes, no change to version number.)
#1. Corrected bug when only one of grm.file or trait.file has been read in and individuals are excluded,
#    which used to result in inconsistent data.
#2. Corrected bug when one of the traits in bivariate analysis was constant. Now produces NAs, previously crashed.
